from plone.app.z3cform.interfaces import ISubmitWidget
from plone.app.z3cform.widgets.base import HTMLInputWidget
from z3c.form.button import ButtonAction as ButtonActionBase
from z3c.form.interfaces import IFieldWidget
from z3c.form.widget import FieldWidget
from z3c.form.widget import Widget
from zope.interface import implementer
from zope.interface import implementer_only


PRIMARY_BUTTON_NAMES = ("add", "save")


@implementer_only(ISubmitWidget)
class SubmitWidget(HTMLInputWidget, Widget):
    """submit widget"""

    klass = "submit-widget btn"

    @property
    def attributes(self):
        attributes = super().attributes
        if self.__name__ not in PRIMARY_BUTTON_NAMES:
            # do disable only primary buttons validation
            attributes["formnovalidate"] = True
        return attributes

    def update(self):
        if self.field.__name__ in PRIMARY_BUTTON_NAMES:
            self.addClass("btn-primary")
        else:
            self.addClass("btn-secondary")


@implementer(IFieldWidget)
def SubmitFieldWidget(field, request):
    submit = FieldWidget(field, SubmitWidget(request))
    submit.value = field.title
    return submit


class ButtonAction(SubmitWidget, ButtonActionBase):
    """need to subclass ButtonAction here
    to get our SubmitWidget adapter
    """



from plone.app.z3cform.interfaces import ICheckBoxWidget
from plone.app.z3cform.widgets.base import HTMLInputWidget
from z3c.form.browser.checkbox import CheckBoxWidget as CheckBoxWidgetBase
from z3c.form.interfaces import IFieldWidget
from z3c.form.widget import FieldWidget
from zope.interface import implementer
from zope.interface import implementer_only


@implementer_only(ICheckBoxWidget)
class CheckBoxWidget(HTMLInputWidget, CheckBoxWidgetBase):
    """Implementation checkbox widget."""

    klass = "checkbox-widget"

    def update(self):
        super().update()
        if self.mode == "input":
            self.addClass("form-check-input")


@implementer(IFieldWidget)
def CheckBoxFieldWidget(field, request):
    return FieldWidget(field, CheckBoxWidget(request))



from plone.app.z3cform.interfaces import IDatetimeWidget
from plone.app.z3cform.interfaces import IDateWidget
from plone.app.z3cform.interfaces import ITimeWidget
from plone.app.z3cform.utils import dict_merge
from plone.app.z3cform.widgets.base import HTMLTextInputWidget
from plone.base import PloneMessageFactory as _
from z3c.form.interfaces import IDataConverter
from z3c.form.interfaces import IFieldWidget
from z3c.form.widget import FieldWidget
from z3c.form.widget import Widget
from zope.component import getMultiAdapter
from zope.i18n import translate
from zope.interface import implementer
from zope.interface import implementer_only


def get_date_options(request):
    calendar = request.locale.dates.calendars["gregorian"]
    return {
        "behavior": "native",
        "week-numbers": "show",
        "first-day": calendar.week.get("firstDay") == 1 and 1 or 0,
        "today": translate(_("Today"), context=request),
        "clear": translate(_("Clear"), context=request),
    }


class DateTimeWidgetBase(HTMLTextInputWidget, Widget):
    _input_type = ""
    _formater = ""
    _formater_length = ""

    default_timezone = None
    default_time = "00:00:00"

    @property
    def attributes(self):
        attributes = super().attributes
        # NOTE: we're setting the "type" attribute here
        # which overrides 'type="text"' in templates/text_input.pt
        attributes["type"] = self._input_type
        return attributes

    def get_pattern_options(self):
        pat_options = dict_merge(
            get_date_options(self.request),
            self.pattern_options,
        )
        if self.field.required:
            pat_options["clear"] = False
        return pat_options

    def render(self):
        """Render widget.

        :returns: Widget's HTML.
        :rtype: string
        """
        if self.mode != "display":
            return super().render()

        if not self.value:
            return ""

        converter = getMultiAdapter((self.field, self), IDataConverter)

        if not converter:
            return self.value

        field_value = converter.toFieldValue(self.value)
        if field_value is self.field.missing_value:
            return ""

        formatter = self.request.locale.dates.getFormatter(
            self._formater,
            self._formater_length,
        )
        return formatter.format(field_value)


@implementer_only(IDateWidget)
class DateWidget(DateTimeWidgetBase):
    """Date widget for z3c.form."""

    _input_type = "date"
    _formater = "date"
    _formater_length = "short"

    pattern = "date-picker"
    klass = "date-widget"


@implementer(IFieldWidget)
def DateFieldWidget(field, request):
    return FieldWidget(field, DateWidget(request))


@implementer_only(IDatetimeWidget)
class DatetimeWidget(DateTimeWidgetBase):
    """Datetime widget for z3c.form."""

    _input_type = "datetime-local"
    _formater = "dateTime"
    _formater_length = "short"

    pattern = "datetime-picker"
    klass = "datetime-widget"


@implementer(IFieldWidget)
def DatetimeFieldWidget(field, request):
    return FieldWidget(field, DatetimeWidget(request))


@implementer_only(ITimeWidget)
class TimeWidget(DateTimeWidgetBase):
    """TimeWidget for z3c.form."""

    _input_type = "time"
    _formater = "time"
    _formater_length = "short"
    klass = "time-widget"

    # no pattern set for time input
    pattern = ""


@implementer(IFieldWidget)
def TimeFieldWidget(field, request):
    return FieldWidget(field, TimeWidget(request))



from plone.app.z3cform.interfaces import IRadioWidget
from plone.app.z3cform.widgets.base import HTMLInputWidget
from z3c.form.browser.radio import RadioWidget as RadioWidgetBase
from z3c.form.interfaces import IFieldWidget
from z3c.form.widget import FieldWidget
from zope.interface import implementer
from zope.interface import implementer_only


@implementer_only(IRadioWidget)
class RadioWidget(HTMLInputWidget, RadioWidgetBase):
    """Implementation of radio widget."""

    klass = "radio-widget"

    def update(self):
        super().update()
        if self.mode == "input":
            self.addClass("form-check-input")


@implementer(IFieldWidget)
def RadioFieldWidget(field, request):
    return FieldWidget(field, RadioWidget(request))



from OFS.interfaces import ISimpleItem
from plone.app.z3cform.interfaces import IAjaxSelectWidget
from plone.app.z3cform.interfaces import ISelect2Widget
from plone.app.z3cform.interfaces import ISelectWidget
from plone.app.z3cform.utils import dict_merge
from plone.app.z3cform.utils import get_context_url
from plone.app.z3cform.utils import get_widget_form
from plone.app.z3cform.widgets.base import HTMLInputWidget
from plone.app.z3cform.widgets.base import HTMLSelectWidget
from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from z3c.form.browser.select import SelectWidget as SelectWidgetBase
from z3c.form.interfaces import IEditForm
from z3c.form.interfaces import IFieldWidget
from z3c.form.interfaces import IForm
from z3c.form.interfaces import ITerms
from z3c.form.interfaces import NO_VALUE
from z3c.form.widget import FieldWidget
from z3c.form.widget import Widget
from zope.component import getUtility
from zope.component import queryUtility
from zope.interface import implementer
from zope.interface import implementer_only
from zope.schema.interfaces import IChoice
from zope.schema.interfaces import ICollection
from zope.schema.interfaces import ISequence
from zope.schema.interfaces import ITreeVocabulary
from zope.schema.interfaces import IVocabularyFactory


@implementer_only(ISelectWidget)
class SelectWidget(HTMLSelectWidget, SelectWidgetBase):
    klass = "select-widget"


@implementer(IFieldWidget)
def SelectFieldWidget(field, request, extra=None):
    if extra is not None:
        request = extra
    return FieldWidget(field, SelectWidget(request))


@implementer(IFieldWidget)
def CollectionChoiceSelectFieldWidget(field, value_type, request):
    """IFieldWidget factory for SelectWidget."""
    return SelectFieldWidget(field, request)


@implementer_only(ISelect2Widget)
class Select2Widget(HTMLSelectWidget, SelectWidgetBase):
    """Select widget for z3c.form."""

    pattern = "select2"
    separator = ";"
    noValueToken = ""
    noValueMessage = ""
    orderable = False

    @property
    def items(self):
        """
        Optionally handle ITreeVocabulary vocabs as dicts.
        """
        terms = self.terms
        if ITerms.providedBy(terms):
            terms = terms.terms

        if ITreeVocabulary.providedBy(terms):
            groups = {}
            for group_term, option_terms in terms.items():
                group_widget = type(self)(self.request)
                group_widget.terms = option_terms
                group_widget.id = self.id
                group_label = group_term.title or group_term.value or group_term.token
                groups[group_label] = super(Select2Widget, group_widget).items
            return groups
        else:
            return super().items

    def get_pattern_options(self):
        pattern_options = {}

        # ISequence represents an orderable collection
        if ISequence.providedBy(self.field) or self.orderable:
            pattern_options["orderable"] = True

        if self.multiple:
            pattern_options["separator"] = self.separator

        # Allow to clear field value if it is not required
        if not self.field.required:
            pattern_options["allowClear"] = True

        return pattern_options

    def extract(self, default=NO_VALUE):
        """Override extract to handle delimited response values.
        Skip the vocabulary validation provided in the parent
        method, since it's not ever done for single selects."""
        if (
            self.name not in self.request
            and self.name + "-empty-marker" in self.request
        ):
            return []
        return self.request.get(self.name, default)


@implementer(IFieldWidget)
def Select2FieldWidget(field, request):
    return FieldWidget(field, Select2Widget(request))


@implementer_only(IAjaxSelectWidget)
class AjaxSelectWidget(HTMLInputWidget, Widget):
    """Ajax select widget for z3c.form."""

    pattern = "select2"
    separator = ";"
    vocabulary = None
    vocabulary_view = "@@getVocabulary"
    orderable = False

    def _view_context(self):
        view_context = get_widget_form(self)
        # For EditForms and non-Forms (in tests), the vocabulary is looked
        # up on the context, otherwise on the view
        if IEditForm.providedBy(view_context):
            if self.is_subform_widget():
                view_context = self.form.parentForm.context
            elif not ISimpleItem.providedBy(self.context):
                view_context = self.form.context
            else:
                view_context = self.context
        elif not IForm.providedBy(view_context):
            view_context = self.context
        return view_context

    def get_vocabulary(self):
        if self.vocabulary and isinstance(self.vocabulary, str):
            factory = queryUtility(
                IVocabularyFactory,
                self.vocabulary,
            )
            if factory:
                return factory(self._view_context())
        return self.vocabulary

    def display_items(self):
        if self.value:
            tokens = self.value.split(self.separator)
            vocabulary = self.get_vocabulary()
            for token in tokens:
                item = {"token": token, "title": token}
                if vocabulary is not None:
                    try:
                        item["title"] = vocabulary.getTermByToken(token).title
                    except LookupError:
                        pass
                yield item

    def has_multiple_values(self):
        return self.value and self.value.split(self.separator)

    def _ajaxselect_options(self):
        options = {
            "separator": self.separator,
        }
        if self.vocabulary:
            options["vocabularyUrl"] = "{}/{}?name={}".format(
                get_context_url(self._view_context()),
                self.vocabulary_view,
                self.vocabulary,
            )
            field_name = self.field and self.field.__name__ or None
            if field_name:
                options["vocabularyUrl"] += f"&field={field_name}"
            vocabulary = self.get_vocabulary()
            if vocabulary is not None and self.value:
                options["initialValues"] = dict()
                for token in self.value.split(self.separator):
                    try:
                        term = vocabulary.getTermByToken(token)
                        options["initialValues"][term.token] = term.title
                    except LookupError:
                        options["initialValues"][token] = token

        return options

    def update(self):
        super().update()
        field = getattr(self, "field", None)
        field = getattr(field, "value_type", field)
        if (
            not self.vocabulary
            and field is not None
            and getattr(field, "vocabularyName", None)
        ):
            self.vocabulary = field.vocabularyName

    def get_pattern_options(self):
        pattern_options = {}
        context = self.context
        field = None

        if IChoice.providedBy(self.field):
            pattern_options["maximumSelectionSize"] = 1
            field = self.field
        elif ICollection.providedBy(self.field):
            field = self.field.value_type
        if IChoice.providedBy(field):
            pattern_options["allowNewItems"] = "false"

        pattern_options = dict_merge(self._ajaxselect_options(), pattern_options)

        if field and getattr(field, "vocabulary", None):
            form_url = self.request.getURL()
            source_url = "{:s}/++widget++{:s}/@@getSource".format(
                form_url,
                self.name,
            )
            pattern_options["vocabularyUrl"] = source_url

        # ISequence represents an orderable collection
        if ISequence.providedBy(self.field) or self.orderable:
            pattern_options["orderable"] = True

        if self.vocabulary == "plone.app.vocabularies.Keywords":
            membership = getToolByName(context, "portal_membership")
            user = membership.getAuthenticatedMember()

            registry = getUtility(IRegistry)
            roles_allowed_to_add_keywords = registry.get(
                "plone.roles_allowed_to_add_keywords", set()
            )
            roles = set(user.getRolesInContext(context))
            allowNewItems = bool(
                roles.intersection(roles_allowed_to_add_keywords),
            )
            pattern_options["allowNewItems"] = str(
                allowNewItems,
            ).lower()

        return pattern_options


@implementer(IFieldWidget)
def AjaxSelectFieldWidget(field, request, extra=None):
    if extra is not None:
        request = extra
    return FieldWidget(field, AjaxSelectWidget(request))



from datetime import date
from datetime import datetime
from datetime import time
from plone.app.z3cform import utils
from plone.app.z3cform.interfaces import IAjaxSelectWidget
from plone.app.z3cform.interfaces import IContentBrowserWidget
from plone.app.z3cform.interfaces import IDatetimeWidget
from plone.app.z3cform.interfaces import IDateWidget
from plone.app.z3cform.interfaces import ILinkWidget
from plone.app.z3cform.interfaces import IQueryStringWidget
from plone.app.z3cform.interfaces import IRelatedItemsWidget
from plone.app.z3cform.interfaces import ISelectWidget
from plone.app.z3cform.interfaces import ISingleCheckBoxBoolWidget
from plone.app.z3cform.interfaces import ITimeWidget
from plone.base.utils import safe_callable
from plone.uuid.interfaces import IUUID
from Products.CMFCore.utils import getToolByName
from z3c.form.converter import BaseDataConverter
from z3c.form.converter import CollectionSequenceDataConverter
from z3c.form.converter import SequenceDataConverter
from z3c.form.interfaces import ISequenceWidget
from z3c.relationfield.interfaces import IRelation
from z3c.relationfield.interfaces import IRelationList
from zope.component import adapter
from zope.component.hooks import getSite
from zope.schema.interfaces import IBool
from zope.schema.interfaces import ICollection
from zope.schema.interfaces import IDate
from zope.schema.interfaces import IDatetime
from zope.schema.interfaces import IField
from zope.schema.interfaces import IList
from zope.schema.interfaces import ITime

import json
import pytz
import urllib


@adapter(IDate, IDateWidget)
class DateWidgetConverter(BaseDataConverter):
    """Data converter for date fields."""

    def toWidgetValue(self, value):
        """Converts from field value to widget.

        :param value: Field value.
        :type value: date

        :returns: Date in format `Y-m-d`
        :rtype: string
        """
        if value is self.field.missing_value:
            return ""
        return ("{value.year:}-{value.month:02}-{value.day:02}").format(value=value)

    def toFieldValue(self, value):
        """Converts from widget value to field.

        :param value: Value inserted by date widget.
        :type value: string

        :returns: `date.date` object.
        :rtype: date
        """
        if not value:
            return self.field.missing_value
        return date(*map(int, value.split("-")))


@adapter(IDatetime, IDatetimeWidget)
class DatetimeWidgetConverter(BaseDataConverter):
    """Data converter for datetime fields."""

    def toWidgetValue(self, value):
        """Converts from field value to widget.

        :param value: Field value.
        :type value: datetime

        :returns: Datetime in format `Y-m-d H:M`
        :rtype: string
        """
        if value is self.field.missing_value:
            return ""
        return (
            f"{value.year:}-{value.month:02}-{value.day:02}T"
            f"{value.hour:02}:{value.minute:02}"
        )

    def toFieldValue(self, value):
        """Converts from widget value to field.

        :param value: Value inserted by datetime widget.
        :type value: string

        :returns: `datetime.datetime` object.
        :rtype: datetime
        """
        if not value:
            return self.field.missing_value
        tmp = value.split("T")
        if not tmp[0]:
            return self.field.missing_value
        value = tmp[0].split("-")
        if len(tmp) == 2 and ":" in tmp[1]:
            value += tmp[1].split(":")
        else:
            default_time = self.widget.default_time
            default_time = (
                default_time(self.widget.context)
                if safe_callable(default_time)
                else default_time
            )
            value += default_time.split(":")

        # TODO: respect the selected zone from the widget and just fall back
        # to default_zone
        default_zone = self.widget.default_timezone
        zone = (
            default_zone(self.widget.context)
            if safe_callable(default_zone)
            else default_zone
        )
        ret = datetime(*map(int, value))
        if zone:
            tzinfo = pytz.timezone(zone)
            ret = tzinfo.localize(ret)
        return ret


@adapter(IDatetime, IDateWidget)
class DateWidgetToDatetimeConverter(BaseDataConverter):
    """Data converter for date widget on datetime fields."""

    def toWidgetValue(self, value):
        """Converts from field value to widget.

        :param value: Field value.
        :type value: datetime

        :returns: Datetime in format `Y-m-d`
        :rtype: string
        """
        if value is self.field.missing_value:
            return ""
        return f"{value.year:}-{value.month:02}-{value.day:02}"

    def toFieldValue(self, value):
        """Converts from widget value to field.

        :param value: Value inserted by datetime widget.
        :type value: string

        :returns: `datetime.datetime` object.
        :rtype: datetime
        """
        if not value:
            return self.field.missing_value
        value = value.split("-")
        if len(value) != 3:
            return self.field.missing_value

        default_time = self.widget.default_time
        default_time = (
            default_time(self.widget.context)
            if safe_callable(default_time)
            else default_time
        )
        value += default_time.split(":")

        # TODO: respect the selected zone from the widget and just fall back
        # to default_zone
        default_zone = self.widget.default_timezone
        zone = (
            default_zone(self.widget.context)
            if safe_callable(default_zone)
            else default_zone
        )
        ret = datetime(*map(int, value))
        if zone:
            tzinfo = pytz.timezone(zone)
            ret = tzinfo.localize(ret)
        return ret


@adapter(ITime, ITimeWidget)
class TimeWidgetConverter(BaseDataConverter):
    """Data converter for datetime fields."""

    type = "time"

    def toWidgetValue(self, value):
        if value == self.field.missing_value:
            return ""
        return value.strftime("%H:%M")

    def toFieldValue(self, value):
        if value == "":
            return self.field.missing_value
        return time(*map(int, value.split(":")))


class Select2WidgetConverterBase:
    def toFieldValue(self, value):
        """Converts from widget value to field.

        :param value: Value inserted by Select2 widget or default html
                      select/multi-select
        :type value: string | list

        :returns: List of items
        :rtype: list | tuple | set
        """
        separator = getattr(self.widget, "separator", ";")
        if isinstance(value, str):
            value = value.strip()
            if value:
                value = value.split(separator)
            else:
                return self.field.missing_value
        elif value == ("",):
            return self.field.missing_value
        return super().toFieldValue(value)


@adapter(IField, ISelectWidget)
class SequenceSelect2WidgetConverter(
    Select2WidgetConverterBase,
    SequenceDataConverter,
):
    """Data converter for IField fields using the Select2Widget."""


@adapter(ICollection, ISelectWidget)
class Select2WidgetConverter(
    Select2WidgetConverterBase,
    CollectionSequenceDataConverter,
):
    """Data converter for ICollection fields using the Select2Widget."""


@adapter(ICollection, IAjaxSelectWidget)
class AjaxSelectWidgetConverter(BaseDataConverter):
    """Data converter for ICollection fields using the AjaxSelectWidget."""

    def toWidgetValue(self, value):
        """Converts from field value to widget tokenized widget value.

        :param value: Field value.
        :type value: list |tuple | set

        :returns: Items separated using separator defined on widget
        :rtype: string
        """
        if not value:
            return ""
        vocabulary = self.widget.get_vocabulary()
        tokenized_value = []
        for term_value in value:
            if vocabulary is not None:
                try:
                    term = vocabulary.getTerm(term_value)
                    tokenized_value.append(term.token)
                    continue
                except (LookupError, ValueError):
                    pass
            tokenized_value.append(str(term_value))
        return getattr(self.widget, "separator", ";").join(tokenized_value)

    def toFieldValue(self, value):
        """Converts from widget value to field.

        :param value: Value inserted by AjaxSelect widget.
        :type value: string

        :returns: List of items
        :rtype: list | tuple | set
        """
        collectionType = self.field._type
        if isinstance(collectionType, tuple):
            collectionType = collectionType[-1]
        if not len(value):
            return self.field.missing_value
        valueType = self.field.value_type._type
        if isinstance(valueType, tuple):
            valueType = valueType[0]
        separator = getattr(self.widget, "separator", ";")
        self.widget.update()  # needed to have a vocabulary
        vocabulary = self.widget.get_vocabulary()
        untokenized_value = []
        for token in value.split(separator):
            if vocabulary is not None:
                try:
                    term = vocabulary.getTermByToken(token)
                    if valueType:
                        untokenized_value.append(valueType(term.value))
                    else:
                        untokenized_value.append(term.value)
                    continue
                except (LookupError, ValueError):
                    pass
            untokenized_value.append(
                valueType(token) if valueType else token,
            )
        return collectionType(untokenized_value)


@adapter(IRelation, IContentBrowserWidget)
class RelationChoiceContentBrowserWidgetConverter(BaseDataConverter):
    """Data converter for RelationChoice fields using the ContentBrowserWidget."""

    def toWidgetValue(self, value):
        if not value:
            return self.field.missing_value
        return IUUID(value)

    def toFieldValue(self, value):
        if not value:
            return self.field.missing_value
        try:
            catalog = getToolByName(self.widget.context, "portal_catalog")
        except AttributeError:
            catalog = getToolByName(getSite(), "portal_catalog")

        res = catalog(UID=value)
        if res:
            return res[0].getObject()
        else:
            return self.field.missing_value


# BBB
@adapter(IRelation, IRelatedItemsWidget)
class RelationChoiceRelatedItemsWidgetConverter(
    RelationChoiceContentBrowserWidgetConverter
):
    """backwards compatibility"""


@adapter(IRelation, ISequenceWidget)
class RelationChoiceSelectWidgetConverter(RelationChoiceContentBrowserWidgetConverter):
    """Data converter for RelationChoice fields using with SequenceWidgets,
    which expect sequence values.
    """

    def toWidgetValue(self, value):
        if not value:
            missing = self.field.missing_value
            return [] if missing is None else missing
        return [IUUID(value)]


@adapter(ICollection, IContentBrowserWidget)
class ContentBrowserDataConverter(BaseDataConverter):
    """Data converter for ICollection fields using the ContentBrowserWidget."""

    def toWidgetValue(self, value):
        """Converts from field value to widget.

        :param value: List of catalog brains.
        :type value: list

        :returns: List of of UID separated by separator defined on widget.
        :rtype: string
        """
        if not value:
            return self.field.missing_value
        separator = getattr(self.widget, "separator", ";")
        if IRelationList.providedBy(self.field):
            return separator.join([IUUID(o) for o in value if o])
        else:
            return separator.join(v for v in value if v)

    def toFieldValue(self, value):
        """Converts from widget value to field.

        :param value: List of UID's separated by separator defined
        :type value: string

        :returns: List of content objects
        :rtype: list | tuple | set
        """
        if not value:
            return self.field.missing_value

        collectionType = self.field._type
        if isinstance(collectionType, tuple):
            collectionType = collectionType[-1]

        separator = getattr(self.widget, "separator", ";")
        # Some widgets (like checkbox) return lists
        if isinstance(value, str):
            value = value.split(separator)

        if IRelationList.providedBy(self.field):
            try:
                catalog = getToolByName(self.widget.context, "portal_catalog")
            except AttributeError:
                catalog = getToolByName(getSite(), "portal_catalog")

            objects = {
                item.UID: item.getObject() for item in catalog(UID=value) if item
            }

            return collectionType(
                objects[uid] for uid in value if uid in objects.keys()
            )
        else:
            valueType = getattr(self.field.value_type, "_type", str)
            if valueType is None:
                valueType = str
            if valueType == bytes:
                return collectionType(valueType(v, encoding="utf8") for v in value)
            return collectionType(valueType(v) for v in value)


# BBB
@adapter(ICollection, IRelatedItemsWidget)
class RelatedItemsDataConverter(ContentBrowserDataConverter):
    """backwards compatibility"""


@adapter(IRelationList, ISequenceWidget)
class RelationListSelectWidgetDataConverter(ContentBrowserDataConverter):
    """Data converter for RelationChoice fields using with SequenceWidgets,
    which expect sequence values.
    """

    def toWidgetValue(self, value):
        """Converts from field value to widget.

        :param value: List of catalog brains.
        :type value: list

        :returns: List of of UID.
        :rtype: list
        """
        if not value:
            missing = self.field.missing_value
            return [] if missing is None else missing
        if IRelationList.providedBy(self.field):
            return [IUUID(o) for o in value if o]
        else:
            return [v for v in value if v]


@adapter(IList, IQueryStringWidget)
class QueryStringDataConverter(BaseDataConverter):
    """Data converter for IList."""

    def toWidgetValue(self, value):
        """Converts from field value to widget.

        :param value: Query string.
        :type value: list

        :returns: Query string converted to JSON.
        :rtype: string
        """
        if not value:
            return "[]"
        return json.dumps(value)

    def toFieldValue(self, value):
        """Converts from widget value to field.

        :param value: Query string.
        :type value: string

        :returns: Query string.
        :rtype: list
        """
        try:
            value = json.loads(value)
        except ValueError:
            value = None
        if not value:
            return self.field.missing_value
        return value


@adapter(IField, ILinkWidget)
class LinkWidgetDataConverter(BaseDataConverter):
    """Data converter for the enhanced link widget."""

    def toWidgetValue(self, value):
        value = super().toWidgetValue(value)
        result = {
            "internal": "",
            "external": "",
            "email": "",
            "email_subject": "",
        }
        if not value:
            return result
        if value.startswith("mailto:"):
            # Handle mail URLs
            value = value[7:]  # strip mailto from beginning
            if "?subject=" in value:
                email, email_subject = value.split("?subject=")
                result["email"] = email
                result["email_subject"] = email_subject
            else:
                result["email"] = value
        else:
            uuid = None
            portal = getSite()
            is_same_domain = utils.is_same_domain(value, portal.absolute_url())
            is_absolute = utils.is_absolute(value)
            if "/resolveuid/" in value and (not is_absolute or is_same_domain):
                # Take the UUID part of a resolveuid url, but only if it's on
                # the same domain.
                uuid = value.rsplit("/", 1)[-1]
            elif not is_absolute or is_absolute and is_same_domain:
                # Handle relative URLs or absolute URLs on the same domain.
                parsed = urllib.parse.urlparse(value)
                if parsed.params or parsed.query or parsed.fragment:
                    # we don't want to loose query parameters
                    # so we don't convert URLs pointing to internal
                    # objects with params, queries or fragments
                    # to uids
                    pass
                else:
                    path = utils.replace_link_variables_by_paths(portal, parsed.path)
                    obj = portal.unrestrictedTraverse(path=path, default=None)
                    if obj is not None:
                        uuid = IUUID(obj, None)
            if uuid is not None:
                result["internal"] = uuid
            else:
                result["external"] = value
        return result

    def toFieldValue(self, value):
        """Converts from widget value to field."""

        if not value:
            return self.field.missing_value
        if isinstance(value, dict):
            internal = value.get("internal")
            external = value.get("external")
            email = value.get("email")
        else:
            return value
        if internal:
            url = "${portal_url}/resolveuid/" + internal
        elif email:
            subject = value.get("email_subject")
            if email[:7] != "mailto:":
                email = "mailto:" + email
            if not subject:
                url = email
            else:
                url = "{email}?subject={subject}".format(
                    email=email,
                    subject=subject,
                )
        else:
            url = external
        return url


@adapter(IBool, ISingleCheckBoxBoolWidget)
class BoolSingleCheckboxDataConverter(BaseDataConverter):
    """Special converter between boolean fields and single checkbox widgets."""

    def toWidgetValue(self, value):
        """Convert from Python bool to token sequence representation."""
        if value:
            return ["selected"]
        return ["unselected"]

    def toFieldValue(self, value):
        """See interfaces.IDataConverter"""
        # consider all different from true as false,
        # this way it works with one checkbox
        return bool(value and value[0] == "selected")



from plone.app.z3cform.interfaces import IPasswordWidget
from plone.app.z3cform.widgets.base import HTMLTextInputWidget
from z3c.form.interfaces import IFieldWidget
from z3c.form.widget import FieldWidget
from z3c.form.widget import Widget
from zope.interface import implementer
from zope.interface import implementer_only


@implementer_only(IPasswordWidget)
class PasswordWidget(HTMLTextInputWidget, Widget):
    """enhanced text widget"""

    klass = "password-widget"


@implementer(IFieldWidget)
def PasswordFieldWidget(field, request):
    return FieldWidget(field, PasswordWidget(request))



from plone.app.z3cform.interfaces import ITextAreaWidget
from plone.app.z3cform.interfaces import ITextLinesWidget
from plone.app.z3cform.interfaces import ITextWidget
from plone.app.z3cform.widgets.base import HTMLTextAreaWidget
from plone.app.z3cform.widgets.base import HTMLTextInputWidget
from z3c.form.interfaces import IFieldWidget
from z3c.form.widget import FieldWidget
from z3c.form.widget import Widget
from zope.interface import implementer
from zope.interface import implementer_only


@implementer_only(ITextWidget)
class TextWidget(HTMLTextInputWidget, Widget):
    """enhanced text widget"""

    klass = "text-widget"


@implementer(IFieldWidget)
def TextFieldWidget(field, request):
    return FieldWidget(field, TextWidget(request))


@implementer_only(ITextAreaWidget)
class TextAreaWidget(HTMLTextAreaWidget, Widget):
    """enhanced textarea widget"""

    klass = "textarea-widget"


@implementer(IFieldWidget)
def TextAreaFieldWidget(field, request):
    return FieldWidget(field, TextAreaWidget(request))


@implementer_only(ITextLinesWidget)
class TextLinesWidget(HTMLTextAreaWidget, Widget):
    """enhanced textarea widget"""

    klass = "textarea-widget"


@implementer(IFieldWidget)
def TextLinesFieldWidget(field, request):
    return FieldWidget(field, TextLinesWidget(request))


@implementer(IFieldWidget)
def TextLinesFieldWidgetFactory(field, value_type, request):
    return TextLinesFieldWidget(field, request)



from Acquisition import aq_base
from plone.app.z3cform.utils import call_callables
from plone.app.z3cform.utils import dict_merge
from z3c.form.browser import widget
from z3c.form.interfaces import IValue
from z3c.form.widget import Widget
from zope.component import queryMultiAdapter
from zope.schema.interfaces import ICollection

import json


class PatternNotImplemented(Exception):
    """Raised when method/property is not implemented"""


class BaseWidget(Widget):
    """Base widget for z3c.form."""

    pattern = None
    pattern_options = {}
    _adapterValueAttributes = Widget._adapterValueAttributes + ("pattern_options",)

    def _base(self, pattern, pattern_options={}):
        """Base widget class."""
        raise PatternNotImplemented

    def _base_args(self):
        """Method which will calculate _base class arguments.

        Returns (as python dictionary):
            - `pattern`: pattern name
            - `pattern_options`: pattern options

        :returns: Arguments which will be passed to _base
        :rtype: dict
        """
        if self.pattern is None:
            raise PatternNotImplemented("'pattern' option is not provided.")
        return {
            "pattern": self.pattern,
            "pattern_options": self.pattern_options.copy(),
        }

    def render(self):
        """Render widget.

        :returns: Widget's HTML.
        :rtype: string
        """
        if self.mode != "input":
            return super().render()

        _base_args = self._base_args()
        _base_args["pattern_options"] = call_callables(
            _base_args["pattern_options"],
            self.context,
        )

        pattern_widget = self._base(**_base_args)
        if getattr(self, "klass", False):
            pattern_widget.klass = "{} {}".format(
                pattern_widget.klass,
                self.klass,
            )
        return pattern_widget.render()

    def is_subform_widget(self):
        return getattr(aq_base(self.form), "parentForm", None) is not None


class PatternFormElement(widget.HTMLFormElement):
    """New implementation of pattern widget with z3c.form extendable attributes"""

    _klass_prefix = "pat-"

    pattern = None
    pattern_options = {}

    def get_pattern_options(self):
        """override this factory to inject the pattern options as
        "data-<self._klass_prefix><self.pattern>" attribute
        """
        return self.pattern_options

    @property
    def attributes(self):
        """add "required" attribute"""

        attributes = super().attributes

        if self.required:
            attributes["required"] = "required"

        if self.pattern:
            # if self.pattern_options is injected include them
            pat_options = dict_merge(
                self.get_pattern_options().copy(),
                self.pattern_options,
            )
            # lookup named multiadapter "pattern_options"
            # explicitly and merge it
            pat_options_adapter = queryMultiAdapter(
                (self.context, self.request, self.form, self.field, self),
                IValue,
                name="pattern_options",
            )
            if pat_options_adapter:
                pat_options = dict_merge(
                    pat_options,
                    pat_options_adapter.get(),
                )
            # if callables are injected resolve them
            pat_options = call_callables(
                pat_options,
                self.context,
            )
            attributes[f"data-{self._klass_prefix}{self.pattern}"] = (
                json.dumps(pat_options) if pat_options else ""
            )

        return attributes

    def update(self):
        super().update()
        if self.pattern:
            self.addClass(f"{self._klass_prefix}{self.pattern}")

    def is_subform_widget(self):
        return getattr(aq_base(self.form), "parentForm", None) is not None


class HTMLInputWidget(PatternFormElement, widget.HTMLInputWidget):
    """InputWidget with pattern options"""


class HTMLTextInputWidget(PatternFormElement, widget.HTMLTextInputWidget):
    """TextInputWidget with pattern options"""

    def update(self):
        super().update()
        if self.mode == "input":
            self.addClass("form-control")


class HTMLTextAreaWidget(PatternFormElement, widget.HTMLTextAreaWidget):
    """TextAreaWidget with pattern options"""

    def update(self):
        super().update()
        if self.mode == "input":
            self.addClass("form-control")


class HTMLSelectWidget(PatternFormElement, widget.HTMLSelectWidget):
    """SelectWidget with pattern options"""

    def update(self):
        super().update()

        if ICollection.providedBy(self.field):
            self.multiple = "multiple"

        if self.mode == "input":
            # if select2 pattern only add "display:block" and not "form-select"
            self.addClass("form-select" if self.pattern != "select2" else "d-block")



from Products.Five.browser import BrowserView
from Products.Five.browser.metaconfigure import ViewMixinForTemplates
from z3c.form.error import ErrorViewTemplateFactory
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile

import os.path
import plone.app.z3cform
import plone.app.z3cform.interfaces
import plone.z3cform.interfaces
import plone.z3cform.templates
import z3c.form.interfaces


def path(filepart):
    return os.path.join(
        os.path.dirname(plone.app.z3cform.__file__),
        "templates",
        filepart,
    )


# Override the layout wrapper view default template with a more Plone-looking
# one


layout_factory = plone.z3cform.templates.ZopeTwoFormTemplateFactory(
    path("layout.pt"),
    form=plone.z3cform.interfaces.IFormWrapper,
    request=plone.app.z3cform.interfaces.IPloneFormLayer,
)

# Override the form for the standard full-page form rendering

form_factory = plone.z3cform.templates.ZopeTwoFormTemplateFactory(
    path("form.pt"),
    form=z3c.form.interfaces.IForm,
    request=plone.app.z3cform.interfaces.IPloneFormLayer,
)


# The ploneform-macros view
class Macros(BrowserView):
    def __getitem__(self, key):
        return self.index.macros[key]


# The widget rendering templates need to be Zope 3 templates
class RenderWidget(ViewMixinForTemplates, BrowserView):
    index = ViewPageTemplateFile("templates/widget.pt")


class RenderContentProvider(ViewMixinForTemplates, BrowserView):
    index = ViewPageTemplateFile("templates/contentprovider-widget.pt")


ErrorViewTemplate = ErrorViewTemplateFactory(
    os.path.join(os.path.dirname(__file__), "templates/error.pt"), "text/html"
)



from zope.deferredimport import deprecated


# do not break
deprecated(
    "Please refactor to plone.app.z3cform.widgets.richtext instead (will be removed in Plone 7)",
    IWysiwygWidget="plone.app.z3cform.interfaces:IRichTextWidget",
    WysiwygFieldWidget="plone.app.z3cform.widgets.richtext:RichTextFieldWidget",
    WysiwygWidget="plone.app.z3cform.widgets.richtext:RichTextWidget",
)



from zope.deferredimport import deprecated


# do not break plone.autoform
deprecated(
    "Use plone.app.z3cform.widgets.richtext.RichTextFieldWidget instead (will be removed in Plone 7)",
    WysiwygFieldWidget="plone.app.z3cform.widgets.richtext:RichTextFieldWidget",
)



from plone.app.textfield.widget import IRichTextWidget as patextfield_IRichTextWidget
from plone.schema.interfaces import IFormLayer
from z3c.form.interfaces import ICheckBoxWidget as ICheckBoxWidgetBase
from z3c.form.interfaces import IOrderedSelectWidget as IOrderedSelectWidgetBase
from z3c.form.interfaces import IRadioWidget as IRadioWidgetBase
from z3c.form.interfaces import ISelectWidget as ISelectWidgetBase
from z3c.form.interfaces import ISingleCheckBoxWidget
from z3c.form.interfaces import ISubmitWidget as ISubmitWidgetBase
from z3c.form.interfaces import ITextAreaWidget as ITextAreaWidgetBase
from z3c.form.interfaces import ITextLinesWidget as ITextLinesWidgetBase
from z3c.form.interfaces import ITextWidget as ITextWidgetBase
from zope.deferredimport import deprecated
from zope.interface import Interface


# this should not be used anymore

deprecated(
    "Use zope.schema.interfaces.IDate instead (will be removed in Plone 7).",
    IDateField="zope.schema.interfaces:IDate",
)

deprecated(
    "Use zope.schema.interfaces.IDatetime instead (will be removed in Plone 7).",
    IDatetimeField="zope.schema.interfaces:IDatetime",
)


class IPloneFormLayer(IFormLayer):
    """Request layer installed via browserlayer.xml"""


class IFieldPermissionChecker(Interface):
    """Adapter factory for checking whether a user has permission to
    edit a specific field on a content object.
    """

    def validate(field_name, vocabulary_name=None):
        """Returns True if the current user has permission to edit the
        `field_name` field.  Returns False if the user does not have
        permission.  Raises and AttributeError if the field cannot be
        found.
        """


class ITextWidget(ITextWidgetBase):
    """Marker interface for base text input"""


class ITextAreaWidget(ITextAreaWidgetBase):
    """Marker interface for base text input"""


class ITextLinesWidget(ITextLinesWidgetBase):
    """Marker interface for base text input"""


class IDateWidget(ITextWidget):
    """Marker interface for the DateWidget."""


class IDatetimeWidget(ITextWidget):
    """Marker interface for the DatetimeWidget."""


class ITimeWidget(ITextWidget):
    """Marker interface for the TimeWidget"""


class IOrderedSelectWidget(IOrderedSelectWidgetBase):
    """Marker interface for the OrderedSelectWidget."""


class ISelectWidget(ISelectWidgetBase):
    """Marker interface for the SelectWidget."""


class ISelect2Widget(ISelectWidget):
    """Marker interface for the Select2Widget."""


class IAjaxSelectWidget(ITextWidget):
    """Marker interface for the AjaxSelectWidget."""


class IQueryStringWidget(ITextWidget):
    """Marker interface for the QueryStringWidget."""


class IRelatedItemsWidget(ITextWidget):
    """Marker interface for the RelatedItemsWidget."""


class IContentBrowserWidget(ITextWidget):
    """Marker interface for the RelatedItemsWidget."""


class IRichTextWidget(patextfield_IRichTextWidget):
    """Marker interface for the TinyMCEWidget."""


class ILinkWidget(ITextWidget):
    """Marker interface for the enhanced link widget."""


class IEmailWidget(ITextWidget):
    """Marker interface for the dumb email widget."""


class IPasswordWidget(ITextWidget):
    """Marker interface for the password widget."""


class ISubmitWidget(ISubmitWidgetBase):
    """Marker interface for SubmitWidget."""


class ICheckBoxWidget(ICheckBoxWidgetBase):
    """Marker for CheckBoxWidget."""


class ISingleCheckBoxBoolWidget(ISingleCheckBoxWidget):
    """Marker interface for the SingleCheckboxBoolWidget."""


class IRadioWidget(IRadioWidgetBase):
    """Radio widget."""

    def renderForValue(value):
        """Render a single radio button element for a given value.

        Here the word ``value`` is used in the HTML sense, in other
        words it is a term token.
        """


class IRichTextWidgetInputModeRenderer(Interface):
    """Marker interface to render multiple wysiwyg editors"""



from lxml import etree
from plone.app.textfield.value import RichTextValue
from plone.app.textfield.widget import RichTextWidget as patext_RichTextWidget
from plone.app.z3cform.interfaces import IRichTextWidget
from plone.app.z3cform.interfaces import IRichTextWidgetInputModeRenderer
from plone.app.z3cform.utils import remove_invalid_xml_characters
from plone.app.z3cform.widgets.base import HTMLTextAreaWidget
from z3c.form.interfaces import IFieldWidget
from z3c.form.widget import FieldWidget
from zope.component import ComponentLookupError
from zope.component import getMultiAdapter
from zope.component import queryUtility
from zope.interface import implementer
from zope.interface import implementer_only

import json
import logging


logger = logging.getLogger(__name__)


def get_tinymce_options(context, field, request):
    """
    We're just going to be looking up settings from
    plone pattern options
    """
    options = {}
    try:
        pattern_options = getMultiAdapter(
            (context, request, field), name="pattern_settings"
        ).tinymce()["data-pat-tinymce"]
        options = json.loads(pattern_options)
    except (ComponentLookupError, AttributeError):
        logger.warning("Can not load tinymce pattern options!", exc_info=True)
    return options


class RichTextWidgetBase(HTMLTextAreaWidget, patext_RichTextWidget):
    @property
    def richtext_value(self):
        return self.value and remove_invalid_xml_characters(self.value.raw) or ""


@implementer_only(IRichTextWidget)
class RichTextWidget(RichTextWidgetBase):
    klass = "richtext-widget"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._pattern = None

    @property
    def pattern(self):
        """dynamically grab the actual pattern name so it will
        work with custom visual editors"""
        if self._pattern is None:
            self._pattern = self.getWysiwygEditor()
        return self._pattern

    def get_pattern_options(self):
        return get_tinymce_options(
            self.wrapped_context(),
            self.field,
            self.request,
        )

    def render(self):
        """Render widget.

        :returns: Widget's HTML.
        :rtype: string
        """
        if self.mode != "display":
            renderer = queryUtility(
                IRichTextWidgetInputModeRenderer,
                name=self.getWysiwygEditor(),
                default=tinymce_richtextwidget_render,
            )
            return renderer(self)

        if not self.value:
            return ""

        if isinstance(self.value, RichTextValue):
            return self.value.output_relative_to(self.context)

        return super().render()

    def render_input_mode(self):
        # MODE "INPUT"
        rendered = ""
        allowed_mime_types = self.allowedMimeTypes()
        if not allowed_mime_types or len(allowed_mime_types) <= 1:
            # Display textarea with default widget
            rendered = super().render()
        else:
            # Let pat-textarea-mimetype-selector choose the widget

            # create a copy of RichTextWidget
            textarea_widget = RichTextWidgetBase(self.request)
            textarea_widget.field = self.field
            textarea_widget.name = self.name
            textarea_widget.value = self.value

            mt_pattern_name = "{}{}".format(
                self._klass_prefix,
                "textareamimetypeselector",
            )

            # Initialize mimetype selector pattern
            # TODO: default_mime_type returns 'text/html', regardless of
            # settings. fix in plone.app.textfield
            value_mime_type = (
                self.value.mimeType if self.value else self.field.default_mime_type
            )
            mt_select = etree.Element("select")
            mt_select.attrib["id"] = f"{self.id}_text_format"
            mt_select.attrib["name"] = f"{self.name}.mimeType"
            mt_select.attrib["class"] = f"form-select {mt_pattern_name}"
            mt_select.attrib[f"data-{mt_pattern_name}"] = json.dumps(
                {
                    "textareaName": self.name,
                    "widgets": {
                        "text/html": {  # TODO: currently, we only support
                            # richtext widget config for
                            # 'text/html', no other mimetypes.
                            "pattern": self.pattern,
                            "patternOptions": self.get_pattern_options(),
                        },
                    },
                },
            )

            # Create a list of allowed mime types
            for mt in allowed_mime_types:
                opt = etree.Element("option")
                opt.attrib["value"] = mt
                if value_mime_type == mt:
                    opt.attrib["selected"] = "selected"
                opt.text = mt
                mt_select.append(opt)

            # Render the combined widget
            textarea_widget.update()
            rendered = "{}\n{}".format(
                textarea_widget.render(),
                etree.tostring(mt_select, encoding="unicode"),
            )
        return rendered


def tinymce_richtextwidget_render(widget):
    return RichTextWidget.render_input_mode(widget)


@implementer(IFieldWidget)
def RichTextFieldWidget(field, request):
    return FieldWidget(field, RichTextWidget(request))



from plone.app.z3cform.interfaces import IOrderedSelectWidget
from plone.app.z3cform.widgets.base import HTMLSelectWidget
from z3c.form.browser.orderedselect import (
    OrderedSelectWidget as OrderedSelectWidgetBase,
)
from z3c.form.interfaces import IFieldWidget
from z3c.form.widget import FieldWidget
from zope.interface import implementer
from zope.interface import implementer_only


@implementer_only(IOrderedSelectWidget)
class OrderedSelectWidget(HTMLSelectWidget, OrderedSelectWidgetBase):
    """Implementation of radio widget."""

    klass = "radio-widget"


@implementer(IFieldWidget)
def OrderedSelectFieldWidget(field, request):
    return FieldWidget(field, OrderedSelectWidget(request))


@implementer(IFieldWidget)
def SequenceChoiceSelectFieldWidget(field, value_type, request):
    """IFieldWidget factory for SelectWidget."""
    return OrderedSelectFieldWidget(field, request)



from plone.app.z3cform.interfaces import IEmailWidget
from plone.app.z3cform.widgets.text import TextWidget
from z3c.form.interfaces import IFieldWidget
from z3c.form.widget import FieldWidget
from zope.interface import implementer
from zope.interface import implementer_only


@implementer_only(IEmailWidget)
class EmailWidget(TextWidget):
    """Implementation of dumb email widget."""

    klass = "email-widget"


@implementer(IFieldWidget)
def EmailFieldWidget(field, request):
    return FieldWidget(field, EmailWidget(request))
